__version__ = "0.0.11"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "flux-hierarchy"
PACKAGE_URL = "https://github.com/converged-computing/flux-hierarchy"
KEYWORDS = "flux, flux-framework, throughput, tree, instances"
DESCRIPTION = "Instance tree generation for organization or higher throughput submission"
LICENSE = "LICENSE"

INSTALL_REQUIRES = (("rich", {"min_version": None}),)

TESTS_REQUIRES = (("pytest", {"min_version": None}),)
INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES
